use std::{fs::OpenOptions, io::Write, sync::Arc, time::Duration};

use anyhow::Result;
use assetpath::AssetPath;
use engine::prelude::{
    cgmath::{vec3, Vector2},
    *,
};

use crate::{
    board::{Board, BoardSlot, BoardSlotState},
    objects::Objects,
    simple_ai::SimpleAI,
};

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
pub enum StoneState {
    ReadyToBePlaced,
    Placed,
    Dead,
}

pub struct Stone {
    pub stone: Entity,
    pub state: StoneState,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
pub enum GameState {
    Waiting,
    Placing,
    Removing,
    Main,
    Won(PlayerColor),
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
pub enum PlayerColor {
    White,
    Black,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
pub enum MillState {
    White,
    Black,
    None,
}

#[derive(Debug, Clone, Copy)]
pub enum LogSeverity {
    Basic,
    Debug,
}

const LOG_SEVERITY: LogSeverity = LogSeverity::Basic;
const LOG_FILE: &str = "millgame.log";
const EXTRA_LOG_FILE: &str = "millgame_extra.log";

impl PlayerColor {
    pub fn swap(&mut self) {
        *self = match *self {
            PlayerColor::White => {
                MillGame::log("swapped to black player", LogSeverity::Debug);

                PlayerColor::Black
            }
            PlayerColor::Black => {
                MillGame::log("swapped to white player", LogSeverity::Debug);

                PlayerColor::White
            }
        }
    }

    pub fn other(&self) -> Self {
        match self {
            PlayerColor::White => PlayerColor::Black,
            PlayerColor::Black => PlayerColor::White,
        }
    }
}

#[derive(Default, Debug)]
struct TurnState {
    white_infos: Vec<(StoneState, Entity, Option<Vector2<f32>>)>,
    black_infos: Vec<(StoneState, Entity, Option<Vector2<f32>>)>,
    board_infos: Vec<((usize, usize, usize), Vector2<f32>, BoardSlotState)>,
}

impl TurnState {
    pub fn new(world: &World, game: &MillGame) -> Result<Self> {
        let mut white_infos = Vec::new();
        let mut black_infos = Vec::new();

        for white_stone in game.white_stones.iter() {
            let pos = match white_stone.state {
                StoneState::ReadyToBePlaced => None,
                StoneState::Placed => Some(
                    world
                        .entity(white_stone.stone)?
                        .get_component::<Location>()?
                        .position()
                        .truncate(),
                ),
                StoneState::Dead => None,
            };

            white_infos.push((white_stone.state, white_stone.stone, pos));
        }

        for black_stone in game.black_stones.iter() {
            let pos = match black_stone.state {
                StoneState::ReadyToBePlaced => None,
                StoneState::Placed => Some(
                    world
                        .entity(black_stone.stone)?
                        .get_component::<Location>()?
                        .position()
                        .truncate(),
                ),
                StoneState::Dead => None,
            };

            black_infos.push((black_stone.state, black_stone.stone, pos));
        }

        let board_infos: Vec<((usize, usize, usize), Vector2<f32>, BoardSlotState)> = game
            .board
            .slots()
            .iter()
            .flatten()
            .flatten()
            .map(|slot| ((slot.x, slot.y, slot.z), slot.position, slot.state()))
            .collect();

        Ok(Self {
            white_infos,
            black_infos,
            board_infos,
        })
    }

    pub fn diff(&self, other: &Self) -> (Self, Self) {
        let mut my_diffs = Self::default();
        let mut other_diffs = Self::default();

        for my_info in self.white_infos.iter() {
            let other_info = other.white_infos.iter().find(|o| my_info.1 == o.1).unwrap();

            if my_info.0 != other_info.0 || my_info.2 != other_info.2 {
                my_diffs.white_infos.push(*my_info);
                other_diffs.white_infos.push(*other_info);
            }
        }

        for my_info in self.black_infos.iter() {
            let other_info = other.black_infos.iter().find(|o| my_info.1 == o.1).unwrap();

            if my_info.0 != other_info.0 || my_info.2 != other_info.2 {
                my_diffs.black_infos.push(*my_info);
                other_diffs.black_infos.push(*other_info);
            }
        }

        for my_slot in self.board_infos.iter() {
            let other_slot = other.board_infos.iter().find(|o| my_slot.0 == o.0).unwrap();

            if my_slot.2 != other_slot.2 {
                my_diffs.board_infos.push(*my_slot);
                other_diffs.board_infos.push(*other_slot);
            }
        }

        (my_diffs, other_diffs)
    }

    pub fn log(&self, severity: LogSeverity) {
        if !self.white_infos.is_empty() {
            MillGame::log(" == WHITE STONES ==", severity);

            for white_info in self.white_infos.iter() {
                MillGame::log(&format!("{:?}", white_info), severity);
            }
        }

        if !self.black_infos.is_empty() {
            MillGame::log(" == BLACK STONES ==", severity);

            for black_info in self.black_infos.iter() {
                MillGame::log(&format!("{:?}", black_info), severity);
            }
        }

        MillGame::log(" == BOARD ==", severity);

        for board_info in self.board_infos.iter() {
            MillGame::log(&format!("{:?}", board_info), severity);
        }
    }

    pub fn is_empty(&self) -> bool {
        self.white_infos.is_empty() && self.black_infos.is_empty() && self.board_infos.is_empty()
    }
}

pub struct MillGame {
    last_turn_timing: Duration,

    board: Board,
    white_stones: [Stone; 9],
    black_stones: [Stone; 9],

    state: GameState,
    current_player: PlayerColor,

    mouse_x: u32,
    mouse_y: u32,

    turn_finished: bool,
    turn_states: Vec<TurnState>,

    selected_field: Option<(usize, usize, usize)>,

    white_player_label: Arc<Label>,
    black_player_label: Arc<Label>,
    _start_button: Arc<Button>,
    _grid: Arc<Grid>,

    _gui: Arc<GuiBuilder>,
}

impl MillGame {
    pub const OFFSET_TO_BOARD: f32 = 0.02;
    pub const TURN_WAIT_TIME: Duration = Duration::from_millis(500);

    pub fn new(world: &mut World) -> Result<()> {
        let board = Board::new(world)?;

        // add light
        let mut sun_light = Light::directional_light(world.resources.get::<Context>().device())?;
        sun_light.set_direction(vec3(10.0, -4.0, -20.0))?;
        sun_light.set_color(vec3(1.0, 1.0, 1.0))?;
        sun_light.set_position(vec3(0.0, 0.0, 100.0))?;
        sun_light.set_power(10000000000.0)?;

        let white_stones = Self::init_nine_stones(world, Color::White)?
            .into_iter()
            .enumerate()
            .map(|(index, mut e)| {
                let location = e.get_component_mut::<Location>()?;
                location.set_position(board.white_start_slots()[index].extend(0.0));

                Ok(Stone {
                    stone: world.add_entity(e)?,
                    state: StoneState::ReadyToBePlaced,
                })
            })
            .collect::<Result<Vec<Stone>>>()?
            .try_into()
            .unwrap_or_else(|_: Vec<Stone>| unreachable!("create array from vec from an array"));

        let black_stones = Self::init_nine_stones(world, Color::try_from("#2c2c2c")?)?
            .into_iter()
            .enumerate()
            .map(|(index, mut e)| {
                let location = e.get_component_mut::<Location>()?;
                location.set_position(board.black_start_slots()[index].extend(0.0));

                Ok(Stone {
                    stone: world.add_entity(e)?,
                    state: StoneState::ReadyToBePlaced,
                })
            })
            .collect::<Result<Vec<Stone>>>()?
            .try_into()
            .unwrap_or_else(|_: Vec<Stone>| unreachable!("create array from vec from an array"));

        let scene = world.resources.get_mut::<Scene>();
        scene.add_light(sun_light)?;

        let view = scene.view_mut();
        view.camera_mut().set_center(board.center());
        view.update_buffer()?;

        let mut camera_control = TopDownCameraControl::new(view)?;
        camera_control.set_zoom_levels(
            (3..60).into_iter().rev().map(|z| z as f32).collect(),
            5,
            view,
        )?;

        world.resources.insert(camera_control);

        let path = AssetPath::from((
            world
                .resources
                .get::<EngineSettings>()
                .resource_base_path
                .as_str(),
            "mainmenu.xml",
        ));
        let gui = GuiBuilder::new(world, &path)?;

        let grid: Arc<Grid> = gui.element("grid")?;
        let white_player_label: Arc<Label> = gui.element("white_player_label")?;
        let black_player_label: Arc<Label> = gui.element("black_player_label")?;
        let start_button: Arc<Button> = gui.element("start")?;

        gui.enable(world)?;

        start_button.set_callback({
            let weak_grid = Arc::downgrade(&grid);

            move |world| {
                if let Some(grid) = weak_grid.upgrade() {
                    grid.detach(world, 1, 0)?;
                }

                let now = world.now();
                let me = world.resources.get_mut::<Self>();

                Self::log(" ===== START GAME =====", LogSeverity::Basic);

                me.state = GameState::Placing;
                me.current_player = PlayerColor::Black;

                me.finish_turn(now);

                Ok(())
            }
        });

        world.resources.insert(SimpleAI::new(PlayerColor::White));

        let me = Self {
            last_turn_timing: Duration::default(),

            board,
            white_stones,
            black_stones,

            state: GameState::Waiting,
            current_player: PlayerColor::White,

            mouse_x: 0,
            mouse_y: 0,

            turn_finished: false,
            turn_states: Vec::new(),

            selected_field: None,

            _grid: grid.clone(),
            white_player_label,
            black_player_label,
            _start_button: start_button.clone(),

            _gui: gui,
        };

        world.resources.insert(me);

        Ok(())
    }

    pub fn finish_turn(&mut self, now: Duration) {
        Self::log(
            &format!("{:?} finished turn", self.current_player),
            LogSeverity::Debug,
        );

        self.last_turn_timing = now;
        self.turn_finished = true;
        self.selected_field = None;
    }

    fn print_game_state(
        &mut self,
        world: &World,
        severity: LogSeverity,
        log_state: bool,
    ) -> Result<()> {
        let current_turn_state = TurnState::new(world, self)?;

        if log_state {
            current_turn_state.log(severity);
        }

        if !self.turn_states.is_empty() {
            let (new_diff, old_diff) = current_turn_state.diff(self.turn_states.last().unwrap());

            // verity that last turn actually something happened
            debug_assert!(!new_diff.is_empty());
            debug_assert!(!old_diff.is_empty());

            Self::log(" ===== OLD DIFF =====", severity);
            old_diff.log(severity);
            Self::log(" ===== NEW DIFF =====", severity);
            new_diff.log(severity);
        }

        self.turn_states.push(current_turn_state);

        Ok(())
    }

    fn next_game_step(&mut self, world: &mut World) -> Result<()> {
        Self::log(
            &format!(
                " =========================== NEW TURN ({:?}) ===========================",
                self.state
            ),
            LogSeverity::Basic,
        );

        self.print_game_state(world, LogSeverity::Basic, false)?;

        match self.state {
            GameState::Placing => {
                // check if all stones are placed
                if !self
                    .black_stones
                    .iter()
                    .any(|stones| stones.state == StoneState::ReadyToBePlaced)
                    && !self
                        .white_stones
                        .iter()
                        .any(|stones| stones.state == StoneState::ReadyToBePlaced)
                {
                    Self::log("change state to Main", LogSeverity::Debug);
                    self.state = GameState::Main;
                }

                self.current_player.swap();
            }
            GameState::Removing => {
                // check if all stones are placed, then decide whether main or placing is next state
                if !self
                    .black_stones
                    .iter()
                    .any(|stones| stones.state == StoneState::ReadyToBePlaced)
                    && !self
                        .white_stones
                        .iter()
                        .any(|stones| stones.state == StoneState::ReadyToBePlaced)
                {
                    Self::log("change state to Main", LogSeverity::Debug);
                    self.state = GameState::Main;
                } else {
                    Self::log("change state to Placing", LogSeverity::Debug);
                    self.state = GameState::Placing;
                }

                self.current_player.swap();
            }
            GameState::Main => {
                // check if each player has enough stones to play
                if self
                    .white_stones
                    .iter()
                    .filter(|stone| stone.state != StoneState::Dead)
                    .collect::<Vec<&Stone>>()
                    .len()
                    <= 3
                {
                    self.state = GameState::Won(PlayerColor::Black);
                    return Ok(());
                }

                if self
                    .black_stones
                    .iter()
                    .filter(|stone| stone.state != StoneState::Dead)
                    .collect::<Vec<&Stone>>()
                    .len()
                    <= 3
                {
                    self.state = GameState::Won(PlayerColor::White);
                    return Ok(());
                }

                self.current_player.swap();
            }
            GameState::Waiting => unreachable!(),
            GameState::Won(_) => (),
        }

        Self::log(
            &format!("current player {:?}", self.current_player),
            LogSeverity::Debug,
        );

        let mut resources = unsafe { remove_life_time_mut(world) }.resources.multi_mut();
        let gui_handler = resources.get::<GuiHandler>();
        let simple_ai = resources.get::<SimpleAI>();

        match self.current_player {
            PlayerColor::White => {
                self.black_player_label
                    .set_background(gui_handler, Color::try_from("#9b9292")?)?;
                self.white_player_label
                    .set_background(gui_handler, Color::Yellow)?;
            }
            PlayerColor::Black => {
                self.white_player_label
                    .set_background(gui_handler, Color::try_from("#9b9292")?)?;
                self.black_player_label
                    .set_background(gui_handler, Color::Yellow)?;
            }
        }

        if self.current_player == simple_ai.player_color {
            Self::log("current player is AI", LogSeverity::Debug);

            simple_ai.step(
                unsafe { remove_life_time_mut(self.stones_mut(simple_ai.player_color)) },
                unsafe { remove_life_time_mut(self.board.slots_mut()) },
                world,
                &mut self.state,
            )?;

            self.finish_turn(world.now());
        }

        Self::log("leave next_game_step", LogSeverity::Debug);

        Ok(())
    }

    fn stones_mut(&mut self, player_color: PlayerColor) -> &mut [Stone; 9] {
        match player_color {
            PlayerColor::White => &mut self.white_stones,
            PlayerColor::Black => &mut self.black_stones,
        }
    }

    fn init_nine_stones(world: &mut World, color: Color) -> Result<[EntityObject; 9]> {
        Ok((0..9)
            .map(|_| Self::init_stone(world, color))
            .collect::<Result<Vec<EntityObject>>>()?
            .try_into()
            .unwrap_or_else(|_: Vec<EntityObject>| {
                unreachable!("create array from vec from an array")
            }))
    }

    fn init_stone(world: &mut World, color: Color) -> Result<EntityObject> {
        let mut marker = AssetHandler::create(world).empty_entity();

        let draw = Draw::new(vec![{
            let mut mesh = AssetMesh::new(
                world.resources.get::<Context>().device(),
                world
                    .resources
                    .get::<EngineSettings>()
                    .graphics_info()?
                    .render_type,
            )?;

            let vertex_buffer = Buffer::builder()
                .set_data(&Objects::create_cylinder(1.1, 0.25, 30))
                .set_usage(VK_BUFFER_USAGE_VERTEX_BUFFER_BIT)
                .set_memory_usage(MemoryUsage::CpuOnly)
                .build(world.resources.get::<Context>().device().clone())?;

            let a: [f32; 3] = color.into();

            mesh.add_primitive(
                vertex_buffer,
                None,
                None,
                PrimitiveMaterial {
                    color: [a[0], a[1], a[2], 1.0],
                    metallic_factor: 0.2,
                    emissive_factor: [0.2, 0.2, 0.2],
                    roughness_factor: 0.8,
                    alpha_mode: AlphaMode::Opaque,
                    alpha_cut_off: 0.5,
                },
                true,
            )?;

            mesh
        }]);
        marker.insert_component(draw);

        Location::new_and_setup(&mut marker)?;

        Ok(marker)
    }

    pub fn place_stone(
        stone: &mut Stone,
        (x, y, z): (usize, usize, usize),
        board: &mut [[[BoardSlot; 3]; 3]; 3],
        player_color: PlayerColor,
        world: &mut World,
        state: &mut GameState,
    ) -> Result<MillState> {
        Self::log("place_stone", LogSeverity::Debug);

        let slot = &mut board[x][y][z];

        match player_color {
            PlayerColor::White => {
                Self::log(
                    &format!(
                        " ==> white stone ({:?}) placed at {:?}",
                        stone.stone,
                        (x, y, z)
                    ),
                    LogSeverity::Basic,
                );

                slot.set_state(BoardSlotState::White(stone.stone));
            }
            PlayerColor::Black => {
                Self::log(
                    &format!(
                        " ==> black stone ({:?}) placed at {:?}",
                        stone.stone,
                        (x, y, z)
                    ),
                    LogSeverity::Basic,
                );

                slot.set_state(BoardSlotState::Black(stone.stone));
            }
        }

        stone.state = StoneState::Placed;

        let entity = world.entity_mut(stone.stone)?;

        let location = entity.get_component_mut::<Location>()?;
        location.set_position(slot.position.extend(0.0));

        let millstate = Self::check_for_mill(&board[x][y][z], board);

        if millstate != MillState::None {
            *state = GameState::Removing;
        }

        Self::log(
            &format!("return place stone with {:?}", millstate),
            LogSeverity::Debug,
        );

        Ok(millstate)
    }

    pub fn move_stone(
        (x, y, z): (usize, usize, usize),
        (tx, ty, tz): (usize, usize, usize),
        world: &mut World,
        board: &mut [[[BoardSlot; 3]; 3]; 3],
        state: &mut GameState,
    ) -> Result<Option<MillState>> {
        Self::log(
            &format!("move stone ({:?}) to ({:?})", (x, y, z), (tx, ty, tz)),
            LogSeverity::Debug,
        );

        let slot = unsafe { remove_life_time_mut(&mut board[x][y][z]) };
        let neighbour = &mut board[tx][ty][tz];

        if neighbour.state() != BoardSlotState::Empty
            || !Self::is_neighbour((x, y, z), (tx, ty, tz))
        {
            Self::log("neighbour not empty", LogSeverity::Debug);
            return Ok(None);
        }

        neighbour.set_state(slot.state());
        slot.set_state(BoardSlotState::Empty);

        world
            .entity_mut(match neighbour.state() {
                BoardSlotState::Black(e) => {
                    Self::log(
                        &format!(
                            " ==> black stone ({:?}) moved from {:?} to {:?}",
                            e,
                            (x, y, z),
                            (tx, ty, tz)
                        ),
                        LogSeverity::Basic,
                    );

                    e
                }
                BoardSlotState::White(e) => {
                    Self::log(
                        &format!(
                            " ==> white stone ({:?}) moved from {:?} to {:?}",
                            e,
                            (x, y, z),
                            (tx, ty, tz)
                        ),
                        LogSeverity::Basic,
                    );

                    e
                }

                _ => unreachable!(),
            })?
            .get_component_mut::<Location>()?
            .set_position(neighbour.position.extend(0.0));

        let millstate = Self::check_for_mill(&board[tx][ty][tz], board);

        if millstate != MillState::None {
            Self::log("mill found!", LogSeverity::Debug);
            *state = GameState::Removing;
        }

        Self::log(
            &format!("return move stone with {:?}", millstate),
            LogSeverity::Debug,
        );

        Ok(Some(millstate))
    }

    pub fn remove_stone(stone: &mut Stone, slot: &mut BoardSlot, world: &mut World) -> Result<()> {
        Self::log("remove_stone", LogSeverity::Debug);

        let entity = match slot.state() {
            BoardSlotState::Black(e) => {
                Self::log(
                    &format!(
                        " ==> white player removes black stone ({:?}) at {:?}",
                        e,
                        (slot.x, slot.y, slot.z)
                    ),
                    LogSeverity::Basic,
                );

                e
            }
            BoardSlotState::White(e) => {
                Self::log(
                    &format!(
                        " ==> black player removes white stone ({:?}) at {:?}",
                        e,
                        (slot.x, slot.y, slot.z)
                    ),
                    LogSeverity::Basic,
                );

                e
            }
            _ => unreachable!(),
        };

        world.remove_entity(entity);

        assert_ne!(stone.state, StoneState::Dead);
        stone.state = StoneState::Dead;

        slot.set_state(BoardSlotState::Empty);

        Ok(())
    }

    fn check_mill(s1: &BoardSlot, s2: &BoardSlot, s3: &BoardSlot) -> MillState {
        if s1.valid() && s2.valid() && s3.valid() {
            if s1.white() && s2.white() && s3.white() {
                return MillState::White;
            } else if s1.black() && s2.black() && s3.black() {
                return MillState::Black;
            }
        }

        MillState::None
    }

    pub fn check_for_mill(slot: &BoardSlot, board: &[[[BoardSlot; 3]; 3]; 3]) -> MillState {
        Self::log("check for mill", LogSeverity::Debug);

        if !(slot.x == 0 && slot.y == 0)
            && !(slot.x == 2 && slot.y == 0)
            && !(slot.x == 0 && slot.y == 2)
            && !(slot.x == 2 && slot.y == 2)
        {
            let state = Self::check_mill(
                &board[slot.x][slot.y][0],
                &board[slot.x][slot.y][1],
                &board[slot.x][slot.y][2],
            );

            if state != MillState::None {
                Self::log(&format!("mill found {:?}", state), LogSeverity::Debug);

                return state;
            }
        }

        let state = Self::check_mill(
            &board[slot.x][0][slot.z],
            &board[slot.x][1][slot.z],
            &board[slot.x][2][slot.z],
        );

        if state != MillState::None {
            Self::log(&format!("mill found {:?}", state), LogSeverity::Debug);

            return state;
        }

        let state = Self::check_mill(
            &board[0][slot.y][slot.z],
            &board[1][slot.y][slot.z],
            &board[2][slot.y][slot.z],
        );

        if state != MillState::None {
            Self::log(&format!("mill found {:?}", state), LogSeverity::Debug);

            return state;
        }

        MillState::None
    }

    fn check_mouse_click<F>(&mut self, scene: &Scene, f: F) -> Result<()>
    where
        F: FnOnce(&mut Self, (usize, usize, usize), &mut [[[BoardSlot; 3]; 3]; 3]) -> Result<()>,
    {
        Self::log("check_mouse_click", LogSeverity::Debug);

        if let Some(world_space) = scene.screen_space_to_world(self.mouse_x, self.mouse_y)? {
            if let Some(slot) = unsafe { remove_life_time(self) }
                .board
                .close_to_marker(world_space.truncate())
            {
                Self::log("click is close to marker", LogSeverity::Debug);

                let slots = unsafe { remove_life_time_mut(self.board.slots_mut()) };

                f(self, (slot.x, slot.y, slot.z), slots)?;
            }
        }

        Ok(())
    }

    fn is_neighbour((x, y, z): (usize, usize, usize), (tx, ty, tz): (usize, usize, usize)) -> bool {
        if (x == tx && y == ty)
            && ((x == 0 && y == 0)
                || (x == 2 && y == 0)
                || (x == 0 && y == 2)
                || (x == 2 && y == 2))
        {
            return false;
        }

        (x as i32 - tx as i32).abs() == 1
            || (y as i32 - ty as i32).abs() == 1
            || (z as i32 - tz as i32).abs() == 1
    }

    pub fn log(s: &str, log_severity: LogSeverity) {
        match (LOG_SEVERITY, log_severity) {
            (LogSeverity::Basic, LogSeverity::Basic) => Self::write(s),
            (LogSeverity::Basic, LogSeverity::Debug) => Self::write_extra(s),
            (LogSeverity::Debug, LogSeverity::Basic) => Self::write(s),
            (LogSeverity::Debug, LogSeverity::Debug) => Self::write(s),
        }
    }

    fn write(s: &str) {
        if let Ok(mut file) = OpenOptions::new().append(true).create(true).open(&LOG_FILE) {
            if let Err(_) = file.write_all(format!("{}\n", s.to_string()).as_bytes()) {}
        }
    }

    fn write_extra(s: &str) {
        if let Ok(mut file) = OpenOptions::new()
            .append(true)
            .create(true)
            .open(EXTRA_LOG_FILE)
        {
            if let Err(_) = file.write_all(format!("{}\n", s.to_string()).as_bytes()) {}
        }
    }

    pub fn update(world: &mut World) -> Result<bool> {
        let now = world.now();
        let me = world.resources.get_mut_unchecked::<Self>();

        if me.turn_finished {
            if (me.last_turn_timing + Self::TURN_WAIT_TIME) < now {
                me.turn_finished = false;
                me.next_game_step(world)?;
            }
        }

        Ok(true)
    }
}

impl EventConsumer for MillGame {
    fn event(&mut self, world: &mut World, event: EngineEvent) -> Result<()> {
        match event {
            EngineEvent::MouseMotion(x, y) => {
                self.mouse_x = x;
                self.mouse_y = y;

                let mut resources = world.resources.multi_mut();
                let camera_controls = resources.get::<TopDownCameraControl>();
                let scene = resources.get::<Scene>();

                camera_controls.mouse_move(x, y, scene.view_mut())?;
            }
            EngineEvent::MouseButtonDown(button) => match button {
                MouseButton::Left => {
                    Self::log(
                        &format!("User click at state {:?}", self.state),
                        LogSeverity::Debug,
                    );

                    match self.state {
                        GameState::Placing => {
                            self.check_mouse_click(
                                world.resources.get_unchecked::<Scene>(),
                                |me, (x, y, z), board| {
                                    if let Some(placable) = unsafe {
                                        remove_life_time_mut(me.stones_mut(me.current_player))
                                    }
                                    .iter_mut()
                                    .find(|stone| stone.state == StoneState::ReadyToBePlaced)
                                    {
                                        if board[x][y][z].is_empty() {
                                            if Self::place_stone(
                                                placable,
                                                (x, y, z),
                                                board,
                                                me.current_player,
                                                world,
                                                &mut me.state,
                                            )? == MillState::None
                                            {
                                                me.finish_turn(world.now());
                                            }
                                        } else {
                                            Self::log(
                                                &format!(
                                                    "slot ({:?}), not empty ({:?})",
                                                    (x, y, z),
                                                    board[x][y][z].state()
                                                ),
                                                LogSeverity::Debug,
                                            );
                                        }
                                    }

                                    Ok(())
                                },
                            )?;
                        }
                        GameState::Removing => {
                            self.check_mouse_click(
                                world.resources.get_unchecked::<Scene>(),
                                |me, (x, y, z), board| {
                                    let slot = &mut board[x][y][z];

                                    Self::log(
                                        &format!(
                                            "state {:?} - player {:?}",
                                            slot.state(),
                                            me.current_player
                                        ),
                                        LogSeverity::Debug,
                                    );

                                    if match (slot.state(), me.current_player) {
                                        (BoardSlotState::Black(_), PlayerColor::White) => true,
                                        (BoardSlotState::White(_), PlayerColor::Black) => true,
                                        _ => false,
                                    } {
                                        let stone = me
                                            .stones_mut(me.current_player.other())
                                            .iter_mut()
                                            .find(|s| {
                                                s.stone
                                                    == match slot.state() {
                                                        BoardSlotState::Black(e) => e,
                                                        BoardSlotState::White(e) => e,
                                                        _ => unreachable!(),
                                                    }
                                            })
                                            .unwrap();

                                        Self::remove_stone(stone, slot, world)?;

                                        me.finish_turn(world.now());
                                    }

                                    Ok(())
                                },
                            )?;
                        }
                        GameState::Main => {
                            self.check_mouse_click(world.resources.get_unchecked::<Scene>() , |me, (tx, ty, tz), board| {
                                match me.selected_field {
                                    Some((x, y, z)) => {
                                        if Self::is_neighbour((x, y, z), (tx, ty, tz)) {
                                            if let Some(millstate)
                                                = Self::move_stone((x,y,z), (tx,ty,tz), world, board, &mut me.state)?  {
                                                if let MillState::None = millstate {
                                                    me.selected_field = None;
                                                    me.finish_turn(world.now());
                                                }
                                            }
                                        }
                                    }
                                    None => {
                                        let slot = &board[tx][ty][tz];

                                        match (
                                            slot.state(),
                                            me.current_player,
                                        ) {
                                            (BoardSlotState::Black(_), PlayerColor::Black)
                                            | (BoardSlotState::White(_), PlayerColor::White) => {
                                                Self::log(
                                                    &format!("Selected ({:?})", (tx, ty, tz)),
                                                    LogSeverity::Basic,
                                                );

                                                    me.selected_field = Some((tx, ty, tz));
                                            }

                                            _ => (),
                                        }
                                    }
                                }

                                Ok(())
                            })?;
                        }
                        GameState::Waiting | GameState::Won(_) => (),
                    }
                }
                MouseButton::Middle => world.resources.get_mut::<TopDownCameraControl>().hold(),
                MouseButton::Right => {
                    if self.selected_field.is_some() {
                        Self::log(
                            &format!("Released selection ({:?})", self.selected_field.unwrap()),
                            LogSeverity::Basic,
                        );

                        self.selected_field = None;
                    }
                }
                MouseButton::Forward => (),
                MouseButton::Backward => (),
            },
            EngineEvent::MouseButtonUp(button) => match button {
                MouseButton::Left => (),
                MouseButton::Middle => world.resources.get_mut::<TopDownCameraControl>().release(),
                MouseButton::Right => (),
                MouseButton::Forward => (),
                MouseButton::Backward => (),
            },
            EngineEvent::MouseWheel(_x, y, direction) => {
                let mut resources = world.resources.multi_mut();
                let camera_controls = resources.get::<TopDownCameraControl>();
                let scene = resources.get::<Scene>();

                camera_controls.scroll_wheel(y, direction, scene.view_mut())?;
            }

            _ => (),
        }

        Ok(())
    }
}
